@@ -1,7 +1,7 @@
 package Perinci::Sub::Complete;
 
-our $DATE = '2015-03-30'; # DATE
-our $VERSION = '0.76'; # VERSION
+our $DATE = '2015-04-02'; # DATE
+our $VERSION = '0.77'; # VERSION
 
 use 5.010001;
 use strict;
@@ -339,8 +339,8 @@ sub complete_arg_val {
 
     # XXX reject if meta's v is not 1.1
 
-    my $args_p = $meta->{args} // {};
-    my $arg_p = $args_p->{$arg} or do {
+    my $args_prop = $meta->{args} // {};
+    my $arg_spec = $args_prop->{$arg} or do {
         $log->tracef("[comp][periscomp] arg '$arg' is not specified in meta, declining");
         goto RETURN_RES;
     };
@@ -348,16 +348,57 @@ sub complete_arg_val {
     my $static;
     eval { # completion sub can die, etc.
 
-        my $comp = $arg_p->{completion};
+        my $comp;
+      GET_COMP_ROUTINE:
+        {
+            $comp = $arg_spec->{completion};
+            if ($comp) {
+                $log->tracef("[comp][periscomp] using arg completion routine from 'completion' property");
+                last GET_COMP_ROUTINE;
+            }
+            my $xcomp = $arg_spec->{'x.completion'};
+            if ($xcomp) {
+                require Module::Path::More;
+                my $mod = "Perinci::Sub::XCompletion::$xcomp->[0]";
+                if (Module::Path::More::module_path(module=>$mod)) {
+                    $log->tracef("[comp][periscomp] loading module %s ...", $mod);
+                    my $mod_pm = $mod; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
+                    require $mod_pm;
+                    my $fref = \&{"$mod\::gen_completion"};
+                    $comp = $fref->(%{ $xcomp->[1] });
+                }
+                if ($comp) {
+                    $log->tracef("[comp][periscomp] using arg completion routine from 'x.completion' attribute");
+                    last GET_COMP_ROUTINE;
+                }
+            }
+            my $ent = $arg_spec->{'x.schema.entity'};
+            if ($ent) {
+                require Module::Path::More;
+                my $mod = "Perinci::Sub::ArgEntity::$ent";
+                if (Module::Path::More::module_path(module=>$mod)) {
+                    $log->tracef("[comp][periscomp] loading module %s ...", $mod);
+                    my $mod_pm = $mod; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
+                    require $mod_pm;
+                    if (defined &{"$mod\::complete_arg_val"}) {
+                        $log->tracef("[comp][periscomp] using arg completion routine from complete_arg_val() from %s", $mod);
+                        $comp = \&{"$mod\::complete_arg_val"};
+                        last GET_COMP_ROUTINE;
+                    }
+                }
+            }
+        } # GET_COMP_ROUTINE
+
         if ($comp) {
             if (ref($comp) eq 'CODE') {
-                $log->tracef("[comp][periscomp] invoking routine specified in arg spec's 'completion' property");
+                $log->tracef("[comp][periscomp] invoking arg completion routine");
                 $fres = $comp->(
                     %$extras,
                     word=>$word, ci=>$ci, arg=>$arg, args=>$args{args});
                 return; # from eval
             } elsif (ref($comp) eq 'ARRAY') {
-                $log->tracef("[comp][periscomp] using array specified in arg spec's 'completion' property: %s", $comp);
+                # this is deprecated but will be supported for some time
+                $log->tracef("[comp][periscomp] using array specified in arg completion routine: %s", $comp);
                 $fres = complete_array_elem(
                     array=>$comp, word=>$word, ci=>$ci);
                 $static++;
@@ -384,26 +425,7 @@ sub complete_arg_val {
             return; # from eval
         }
 
-        my $ent = $arg_p->{'x.schema.entity'};
-        if ($ent) {
-            require Module::Path::More;
-            my $mod = "Perinci::Sub::ArgEntity::$ent";
-            if (Module::Path::More::module_path(module=>$mod)) {
-                $log->tracef("[comp][periscomp] loading module %s ...", $mod);
-                my $mod_pm = $mod; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
-                require $mod_pm;
-                if (defined &{"$mod\::complete_arg_val"}) {
-                    $log->tracef("[comp][periscomp] invoking complete_arg_val from %s ...", $mod);
-                    my $fref = \&{"$mod\::complete_arg_val"};
-                    $fres = $fref->(
-                        %$extras,
-                        word=>$word, ci=>$ci, arg=>$arg, args=>$args{args});
-                    return; # from eval
-                }
-            }
-        }
-
-        my $sch = $arg_p->{schema};
+        my $sch = $arg_spec->{schema};
         unless ($sch) {
             $log->tracef("[comp][periscomp] arg spec does not specify schema, declining");
             return; # from eval
@@ -470,8 +492,8 @@ sub complete_arg_elem {
 
     # XXX reject if meta's v is not 1.1
 
-    my $args_p = $meta->{args} // {};
-    my $arg_p = $args_p->{$arg} or do {
+    my $args_prop = $meta->{args} // {};
+    my $arg_spec = $args_prop->{$arg} or do {
         $log->tracef("[comp][periscomp] arg '$arg' is not specified in meta, declining");
         goto RETURN_RES;
     };
@@ -479,18 +501,58 @@ sub complete_arg_elem {
     my $static;
     eval { # completion sub can die, etc.
 
-        my $elcomp = $arg_p->{element_completion};
+        my $elcomp;
+      GET_ELCOMP_ROUTINE:
+        {
+            $elcomp = $arg_spec->{element_completion};
+            if ($elcomp) {
+                $log->tracef("[comp][periscomp] using arg element completion routine from 'element_completion' property");
+                last GET_ELCOMP_ROUTINE;
+            }
+            my $xelcomp = $arg_spec->{'x.element_completion'};
+            if ($xelcomp) {
+               require Module::Path::More;
+                my $mod = "Perinci::Sub::XCompletion::$xelcomp->[0]";
+                if (Module::Path::More::module_path(module=>$mod)) {
+                    $log->tracef("[comp][periscomp] loading module %s ...", $mod);
+                    my $mod_pm = $mod; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
+                    require $mod_pm;
+                    my $fref = \&{"$mod\::gen_completion"};
+                    $elcomp = $fref->(%{ $xelcomp->[1] });
+                }
+                if ($elcomp) {
+                    $log->tracef("[comp][periscomp] using arg element completion routine from 'x.element_completion' attribute");
+                    last GET_ELCOMP_ROUTINE;
+                }
+            }
+            my $ent = $arg_spec->{'x.schema.element_entity'};
+            if ($ent) {
+                require Module::Path::More;
+                my $mod = "Perinci::Sub::ArgEntity::$ent";
+                if (Module::Path::More::module_path(module=>$mod)) {
+                    $log->tracef("[comp][periscomp] loading module %s ...", $mod);
+                    my $mod_pm = $mod; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
+                    require $mod_pm;
+                    if (defined &{"$mod\::complete_arg_val"}) {
+                        $log->tracef("[comp][periscomp] using arg element completion routine from complete_arg_val() from %s", $mod);
+                        $elcomp = \&{"$mod\::complete_arg_val"};
+                        last GET_ELCOMP_ROUTINE;
+                    }
+                }
+            }
+        } # GET_ELCOMP_ROUTINE
+
         $ourextras->{index} = $index;
         if ($elcomp) {
             if (ref($elcomp) eq 'CODE') {
-                $log->tracef("[comp][periscomp] invoking routine specified in arg spec's 'element_completion' property");
+                $log->tracef("[comp][periscomp] invoking arg element completion routine");
                 $fres = $elcomp->(
                     %$extras,
                     %$ourextras,
                     word=>$word, ci=>$ci);
                 return; # from eval
             } elsif (ref($elcomp) eq 'ARRAY') {
-                $log->tracef("[comp][periscomp] using array specified in arg spec's 'element_completion' property: %s", $elcomp);
+                $log->tracef("[comp][periscomp] using array specified in arg element completion routine: %s", $elcomp);
                 $fres = complete_array_elem(
                     array=>$elcomp, word=>$word, ci=>$ci);
                 $static = $word eq '';
@@ -518,26 +580,7 @@ sub complete_arg_elem {
             return; # from eval
         }
 
-        my $ent = $arg_p->{'x.schema.element_entity'};
-        if ($ent) {
-            require Module::Path::More;
-            my $mod = "Perinci::Sub::ArgEntity::$ent";
-            if (Module::Path::More::module_path(module=>$mod)) {
-                $log->tracef("[comp][periscomp] loading module %s ...", $mod);
-                my $mod_pm = $mod; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
-                require $mod_pm;
-                if (defined &{"$mod\::complete_arg_val"}) {
-                    $log->tracef("[comp][periscomp] invoking complete_arg_val from %s ...", $mod);
-                    my $fref = \&{"$mod\::complete_arg_val"};
-                    $fres = $fref->(
-                        %$extras,
-                        word=>$word, ci=>$ci, arg=>$arg, args=>$args{args});
-                    return; # from eval
-                }
-            }
-        }
-
-        my $sch = $arg_p->{schema};
+        my $sch = $arg_spec->{schema};
         unless ($sch) {
             $log->tracef("[comp][periscomp] arg spec does not specify schema, declining");
             return; # from eval
@@ -705,7 +748,7 @@ sub complete_cli_arg {
     my $fres;
 
     my $word   = $words->[$cword];
-    my $args_p = $meta->{args} // {};
+    my $args_prop = $meta->{args} // {};
 
     $log->tracef('[comp][periscomp] entering %s(), words=%s, cword=%d, word=<%s>',
                  $fname, $words, $cword, $word);
@@ -752,7 +795,7 @@ sub complete_cli_arg {
             if ($sm->{arg}) {
                 $log->tracef("[comp][periscomp] completing option value for a known function argument, arg=<%s>, ospec=<%s>", $sm->{arg}, $ospec);
                 $cargs{arg} = $sm->{arg};
-                my $as = $args_p->{$sm->{arg}} or goto RETURN_RES;
+                my $arg_spec = $args_prop->{$sm->{arg}} or goto RETURN_RES;
                 if ($comp) {
                     $log->tracef("[comp][periscomp] invoking routine supplied from 'completion' argument");
                     my $compres;
@@ -820,10 +863,10 @@ sub complete_cli_arg {
             my $fasa = $args{func_arg_starts_at} // 0;
 
             # find if there is a non-greedy argument with the exact position
-            for my $an (keys %$args_p) {
-                my $as = $args_p->{$an};
-                next unless !$as->{greedy} &&
-                    defined($as->{pos}) && $as->{pos} == $pos - $fasa;
+            for my $an (keys %$args_prop) {
+                my $arg_spec = $args_prop->{$an};
+                next unless !$arg_spec->{greedy} &&
+                    defined($arg_spec->{pos}) && $arg_spec->{pos} == $pos - $fasa;
                 $log->tracef("[comp][periscomp] this argument position is for non-greedy function argument <%s>", $an);
                 $cargs{arg} = $an;
                 if ($comp) {
@@ -845,12 +888,12 @@ sub complete_cli_arg {
             # find if there is a greedy argument which takes elements at that
             # position
             for my $an (sort {
-                ($args_p->{$b}{pos} // 9999) <=> ($args_p->{$a}{pos} // 9999)
-            } keys %$args_p) {
-                my $as = $args_p->{$an};
-                next unless $as->{greedy} &&
-                    defined($as->{pos}) && $as->{pos} <= $pos - $fasa;
-                my $index = $pos - $fasa - $as->{pos};
+                ($args_prop->{$b}{pos} // 9999) <=> ($args_prop->{$a}{pos} // 9999)
+            } keys %$args_prop) {
+                my $arg_spec = $args_prop->{$an};
+                next unless $arg_spec->{greedy} &&
+                    defined($arg_spec->{pos}) && $arg_spec->{pos} <= $pos - $fasa;
+                my $index = $pos - $fasa - $arg_spec->{pos};
                 $cargs{arg} = $an;
                 $cargs{index} = $index;
                 $log->tracef("[comp][periscomp] this position is for greedy function argument <%s>'s element[%d]", $an, $index);
@@ -921,7 +964,7 @@ Perinci::Sub::Complete - Complete command-line argument using Rinci metadata
 
 =head1 VERSION
 
-This document describes version 0.76 of Perinci::Sub::Complete (from Perl distribution Perinci-Sub-Complete), released on 2015-03-30.
+This document describes version 0.77 of Perinci::Sub::Complete (from Perl distribution Perinci-Sub-Complete), released on 2015-04-02.
 
 =head1 SYNOPSIS
 
@@ -1301,7 +1344,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Perinci-Su
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/sharyanto/perl-Perinci-Sub-Complete>.
+Source repository is at L<https://github.com/perlancar/perl-Perinci-Sub-Complete>.
 
 =head1 BUGS
 
